from django.db.models import Model, CharField, PositiveSmallIntegerField, Index, Manager, Count


class FakeDataManager(Manager):
    def provider_count(self, provider):
        return self.filter(provider=provider).values('provider').annotate(count=Count('provider')).values('count')


class FakeData(Model):
    objects = FakeDataManager()

    provider = CharField(max_length=255, verbose_name="Faker provider", db_index=True)
    provider_offset = PositiveSmallIntegerField()
    content = CharField(max_length=255, verbose_name="Fake content")

    class Meta(object):
        indexes = [
            Index(fields=['provider', 'provider_offset']),
        ]
        unique_together = (('provider', 'provider_offset'),)

    def __str__(self):
        return u"{0}: '{1}'".format(self.provider, self.content)



# -*- coding: utf-8
from __future__ import unicode_literals, absolute_import

import django

DEBUG = True
USE_TZ = True

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "hhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhh"

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}

ROOT_URLCONF = "tests.urls"

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sites",
    "django_scrubber",
]

SITE_ID = 1

if django.VERSION >= (1, 10):
    MIDDLEWARE = ()
else:
    MIDDLEWARE_CLASSES = ()



from __future__ import absolute_import

import logging

import faker

from django.db.models import Func, Subquery, OuterRef
from django.db.utils import IntegrityError
from django.utils.translation import to_locale, get_language

from . import ScrubberInitError, settings_with_fallback
from .models import FakeData

logger = logging.getLogger(__name__)


class Hash(Func):
    function = 'MD5'
    arity = 1


class Lorem(Func):
    arity = 0
    template = (
        "'Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore "
        "et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut "
        "aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse "
        "cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in "
        "culpa qui officia deserunt mollit anim id est laborum.'")


class Faker(object):
    INITIALIZED_PROVIDERS = set()

    def __init__(self, provider):
        self.provider = provider

    def _initialize_data(self):
        faker_instance = faker.Faker(locale=to_locale(get_language()))

        logger.info('Initializing fake scrub data for provider %s' % self.provider)
        # TODO: maybe be a bit smarter and only regenerate if needed?
        FakeData.objects.filter(provider=self.provider).delete()
        fakedata = []

        # if we don't reset the seed for each provider, registering a new one might change all
        # data for subsequent providers
        faker_instance.seed(settings_with_fallback('SCRUBBER_RANDOM_SEED'))
        for i in range(settings_with_fallback('SCRUBBER_ENTRIES_PER_PROVIDER')):
            fakedata.append(FakeData(provider=self.provider, provider_offset=i,
                                     content=faker_instance.format(self.provider)))

        try:
            FakeData.objects.bulk_create(fakedata)
        except IntegrityError as e:
            raise ScrubberInitError('Integrity error initializing faker data (%s); maybe decrease '
                                    'SCRUBBER_ENTRIES_PER_PROVIDER?' % (e,))

        self.INITIALIZED_PROVIDERS.add(self.provider)

    def __call__(self, *args, **kwargs):
        if self.provider not in self.INITIALIZED_PROVIDERS:
            self._initialize_data()

        # import it here to enable global scrubbers in settings.py
        from .models import FakeData

        return Subquery(FakeData.objects.filter(
            provider=self.provider,
            provider_offset=OuterRef('mod_pk')  # this outer field gets annotated before .update()
            # TODO: This can be used instead of the annotated mod_pk, as soon as this issue is fixed:
            # https://code.djangoproject.com/ticket/28621
            # This would allow us to have per-provider cardinality.
            # provider_offset=Mod(OuterRef('pk'), Subquery(FakeData.objects.provider_count(OuterRef('provider'))))
        ).values('content')[:1])



from __future__ import absolute_import

import logging

from django.conf import settings
from django.db.models import F
from django.db.utils import IntegrityError
from django.core.management.base import BaseCommand, CommandError
from django.apps import apps

from ... import settings_with_fallback

logger = logging.getLogger(__name__)

# py2/3 compat; no need for six
try:
    range = xrange
except NameError:
    range = range


class Command(BaseCommand):
    help = 'Replace database data according to model-specific or global scrubbing rules.'
    leave_locale_alone = True

    def handle(self, *args, **kwargs):
        if not settings.DEBUG:
            # avoid logger, otherwise we might silently fail if we're on live and logging is being sent somewhere else
            self.stderr.write('this command should only be run with DEBUG=True, to avoid running on live systems')
            return False

        global_scrubbers = settings_with_fallback('SCRUBBER_GLOBAL_SCRUBBERS')
        for model in apps.get_models():
            if settings_with_fallback('SCRUBBER_SKIP_UNMANAGED') and not model._meta.managed:
                continue
            if (settings_with_fallback('SCRUBBER_APPS_LIST') and
                    model._meta.app_config.name not in settings_with_fallback('SCRUBBER_APPS_LIST')):
                continue

            scrubbers = dict()
            for field in model._meta.fields:
                if field.name in global_scrubbers:
                    scrubbers[field.name] = global_scrubbers[field.name]
                elif type(field) in global_scrubbers:
                    scrubbers[field.name] = global_scrubbers[type(field)]

            try:
                scrubbers.update(_get_fields(getattr(model, 'Scrubbers')))
            except AttributeError:
                pass  # no model-specific scrubbers

            if not scrubbers:
                continue

            realized_scrubbers = _filter_out_disabled(_call_callables(scrubbers))

            logger.info('Scrubbing %s with %s', model._meta.label, realized_scrubbers)

            try:
                model.objects.annotate(
                    mod_pk=F('pk') % settings_with_fallback('SCRUBBER_ENTRIES_PER_PROVIDER')
                    ).update(**realized_scrubbers)
            except IntegrityError as e:
                raise CommandError('Integrity error while scrubbing %s (%s); maybe increase '
                                   'SCRUBBER_ENTRIES_PER_PROVIDER?' % (model, e))


def _call_callables(d):
    '''
    Helper to realize lazy scrubbers, like Faker, or global field-type scrubbers
    '''
    return {k: (callable(v) and v(k) or v) for k, v in d.items()}


def _get_fields(d):
    '''
    Helper to get "normal" (i.e.: non-magic and non-dunder) instance attributes
    '''
    return {k: v for k, v in vars(d).items() if not k.startswith('_')}


def _filter_out_disabled(d):
    '''
    Helper to remove Nones (actually any false-like type) from the scrubbers.
    This is needed so we can disable global scrubbers in a per-model basis.
    '''
    return {k: v for k, v in d.items() if v}



from django.conf import settings

__version__ = '0.1.0'

defaults = {
    'SCRUBBER_RANDOM_SEED': 42,  # we prefer idempotent scrubbing
    'SCRUBBER_ENTRIES_PER_PROVIDER': 1000,
    'SCRUBBER_GLOBAL_SCRUBBERS': {},
    'SCRUBBER_SKIP_UNMANAGED': True,
    'SCRUBBER_APPS_LIST': None,
}


# can be replaced with ChainMap if we only support py3
def settings_with_fallback(key):
    return getattr(settings, key, defaults[key])


class ScrubberInitError(Exception):
    pass



# -*- coding: utf-8 -*-
# Generated by Django 1.11.9 on 2018-01-24 08:52
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='FakeData',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('provider', models.CharField(db_index=True, max_length=255, verbose_name='Faker provider')),
                ('provider_offset', models.PositiveSmallIntegerField()),
                ('content', models.CharField(max_length=255, verbose_name='Fake content')),
            ],
        ),
        migrations.AddIndex(
            model_name='fakedata',
            index=models.Index(fields=['provider', 'provider_offset'], name='django_scru_provide_d7f250_idx'),
        ),
        migrations.AlterUniqueTogether(
            name='fakedata',
            unique_together=set([('provider', 'provider_offset')]),
        ),
    ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_django_scrubber
------------

Tests for `django_scrubber` models module.
"""

from django.test import TestCase
from django.db.utils import IntegrityError

from django_scrubber import models


class TestDjango_scrubber(TestCase):

    def test_uniqueness(self):
        models.FakeData.objects.create(provider='foo', provider_offset=0, content='bar')
        with self.assertRaises(IntegrityError):
            models.FakeData.objects.create(provider='foo', provider_offset=0, content='baz')
